/*
	*
	* Copyright 2019 Britanicus <marcusbritanicus@gmail.com>
	*
	* This file is a part of DesQ project (https://gitlab.com/desq/)
	*

	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* (at your option) any later version.
	*

	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	* GNU General Public License for more details.
	*

	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, write to the Free Software
	* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	* MA 02110-1301, USA.
	*
*/

// Local Headers
#include "Global.hpp"
#include "DesQDock.hpp"

#include <desq/DesQUtils.hpp>
#include <desqwl/DesQWlGlobal.hpp>
#include <desqwl/Registry.hpp>
#include <desqwl/AutoHideWindow.hpp>

#include <qpa/qplatformnativeinterface.h>

#include "wayland-wlr-layer-shell-unstable-v1-client-protocol.h"
#include "wayland-xdg-shell-client-protocol.h"

struct wl_display *display;
struct wl_registry *registry;
static struct zwlr_layer_shell_v1 *layer_shell;
struct zwlr_layer_surface_v1 *layer_surface;
struct wl_surface *wl_surface;

struct xdg_wm_base *base;
struct xdg_positioner *positioner;

// DesQ Wayland Classes
DesQHotSpot *dockHS;
WaylandRegistry *wlReg;

QPlatformNativeInterface *native;
DesQDock *dock;

uint32_t layer = ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY;
uint32_t anchor = ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM;

static void handle_global(void *data, struct wl_registry *registry, uint32_t name, const char *interface, uint32_t version) {
	if ( strcmp( interface, xdg_wm_base_interface.name ) == 0 )
		base = (xdg_wm_base*)wl_registry_bind( registry, name, &xdg_wm_base_interface, 1 );

	else if ( strcmp( interface, zwlr_layer_shell_v1_interface.name ) == 0 )
		layer_shell = (zwlr_layer_shell_v1*)wl_registry_bind( registry, name, &zwlr_layer_shell_v1_interface, 1 );
}

static void handle_global_remove(void *data, struct wl_registry *registry, uint32_t name) {
	// who cares
}

static const struct wl_registry_listener registry_listener = {
	.global = handle_global,
	.global_remove = handle_global_remove,
};

static void layer_surface_configure(void *data, struct zwlr_layer_surface_v1 *surface, uint32_t serial, uint32_t w, uint32_t h) {

	if ( dock )
		dock->resize( w, h );

	zwlr_layer_surface_v1_ack_configure( surface, serial );
};

static void layer_surface_closed( void *data, struct zwlr_layer_surface_v1 *surface ) {

	zwlr_layer_surface_v1_destroy( surface );
};

struct zwlr_layer_surface_v1_listener layer_surface_listener = {
	.configure = layer_surface_configure,
	.closed = layer_surface_closed,
};

static void xdg_surface_configure_cb( void *data, struct xdg_surface *surface, uint32_t serial ) {

	xdg_surface_set_window_geometry( surface, 50, 50, 800, 600 );
	xdg_surface_ack_configure( surface, serial );
};

struct xdg_surface_listener xdg_surface_listener_par = {
	.configure = xdg_surface_configure_cb
};

struct xdg_surface_listener xdg_surface_listener_pop = {
	.configure = xdg_surface_configure_cb
};

static void xdg_popup_configure_cb( void *data, struct xdg_popup *popup, int32_t x, int32_t y, int32_t width, int32_t height ) {

};

static void xdg_popup_close_cb( void *data, struct xdg_popup *surface ) {

	xdg_popup_destroy( surface );
};

struct xdg_popup_listener xdg_popup_listener = {
	.configure = xdg_popup_configure_cb,
	.popup_done = xdg_popup_close_cb,
};

static void xdg_toplevel_configure_cb(void *data, struct xdg_toplevel *xdg_toplevel, int32_t width, int32_t height, struct wl_array *states) {

};

static void xdg_toplevel_close_cb( void *data, struct xdg_toplevel *surface ) {

	xdg_toplevel_destroy( surface );
};

struct xdg_toplevel_listener xdg_toplevel_listener = {
	.configure = xdg_toplevel_configure_cb,
	.close = xdg_toplevel_close_cb,
};

static void show() {

	qputenv( "QT_WAYLAND_SHELL_INTEGRATION", "desq-dock" );
	dock->show();

	/* Create a layer surface */
	wl_surface = static_cast<struct wl_surface *>( native->nativeResourceForWindow( "surface", dock->windowHandle() ) );
	layer_surface = zwlr_layer_shell_v1_get_layer_surface( layer_shell, wl_surface, NULL, layer, "Dock" );
	zwlr_layer_surface_v1_set_margin( layer_surface, 0, 0, 5, 0 );
	zwlr_layer_surface_v1_set_size( layer_surface, dock->width(), dock->height() );
	zwlr_layer_surface_v1_set_anchor( layer_surface, anchor );
	zwlr_layer_surface_v1_set_exclusive_zone( layer_surface, -1 );
	zwlr_layer_surface_v1_set_keyboard_interactivity( layer_surface, 0 );
	zwlr_layer_surface_v1_add_listener( layer_surface, &layer_surface_listener, layer_surface );
	wl_surface_commit( wl_surface );
	wl_display_roundtrip( display );

	/* Hack to disable wayland integration */
	qunsetenv( "QT_WAYLAND_SHELL_INTEGRATION" );
};

void showMenu() {

	QMainWindow *mw = new QMainWindow();
	mw->setFixedSize( 800, 600 );
	mw->show();

	struct wl_surface *wl_par_surf = static_cast<struct wl_surface *>( native->nativeResourceForWindow( "surface", mw->windowHandle() ) );

	xdg_surface *parent = xdg_wm_base_get_xdg_surface( base, wl_par_surf );
	xdg_surface_add_listener( parent, &xdg_surface_listener_par, parent );

	xdg_toplevel *mwtl = xdg_surface_get_toplevel( parent );
	xdg_toplevel_add_listener( mwtl, &xdg_toplevel_listener, mwtl );
	xdg_toplevel_set_title( mwtl, "DesQ Dock" );
	xdg_toplevel_set_app_id( mwtl, "desq-term" );

	wl_surface_commit( wl_par_surf );
	wl_display_roundtrip( display );
};

int main( int argc, char **argv ) {

	// qputenv( "QT_WAYLAND_DISABLE_DECORATION", "1" );
	qputenv( "QT_WAYLAND_DISABLE_WINDOWDECORATION", "1" );

	qInstallMessageHandler( Logger );
	QCoreApplication::setAttribute( Qt::AA_EnableHighDpiScaling );

	QApplication app( argc, argv );

	dock = new DesQDock();
	QObject::connect( dock, &DesQDock::newWindow, [=]() { showMenu(); } );

	native = QGuiApplication::platformNativeInterface();
	display = ( struct wl_display * )native->nativeResourceForWindow( "display", NULL );
	registry = wl_display_get_registry( display );
	wl_registry_add_listener( registry, &registry_listener, NULL );
	wl_display_roundtrip( display );

	wlReg = new WaylandRegistry( display );
	wlReg->initialize();

	dockHS = new DesQHotSpot( qApp->primaryScreen(), wlReg->wayfireShell(), dock );
	dockHS->createHotspot( DesQHotSpot::Bottom, 1, 500 );
	QObject::connect( dockHS, &DesQHotSpot::enteredHotspot, [=]() { show(); } );

	return app.exec();
};
